package tsp.csc4509.tpnote;

import static tsp.csc4509.common.Log.COMM;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.Serializable;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import tsp.csc4509.spool.Information;
import tsp.csc4509.spool.Spool;
import tsp.csc4509.tcpnio.FullDuplexMsgWorker;
import tsp.csc4509.tcpnio.ReadMessageStatus;

/**
 * Classe du client.
 * 
 * @author
 *
 */
public class Client implements SenderReceiver {
	/**
	 * Taille des buffer.
	 */
	private static final int BUFFSIZE = 1024;

	/**
	 * répertoire ou seront stocké les spool des clients. Pour un client identifié
	 * par l'id userId, le répertoire de son sppol sera: CLIENTSSPOOLDIR/userId
	 */
	public static final String CLIENTSSPOOLDIR = "data/spool/";

	/**
	 * message d'erreur affiché en cas d'erreut de protocole.
	 */
	private static final String BADMESSAGE = "Bad message received from server.";

	/**
	 * Identifiant du client géré.
	 */
	private String userId;

	/**
	 * worker ouvert pour les échanges des messages du protocole.
	 */
	private FullDuplexMsgWorker worker;

	/**
	 * host du serveur. Mémorisé dans les attributs, car il est utilisé lors des
	 * connexions établies pour l'échange du contenu des fichiers.
	 */
	private String serverHost;

	/**
	 * Contructeur du client.
	 * 
	 * @param userId     Identifiant du client géré.
	 * @param serverHost host du serveur/
	 * @param serverPort port pour la connexion qui servira aux échanges des
	 *                   messages du protocole.
	 * @throws IOException toutes les exceptions qui peuvent se produire à
	 *                     l'établissement de la connexion et l'initialisation du
	 *                     spool du client.
	 */
	public Client(final String userId, final String serverHost, final int serverPort) throws IOException {
		super();
		this.serverHost = serverHost;
		worker = new FullDuplexMsgWorker(SocketChannel.open(new InetSocketAddress(serverHost, serverPort)));
		this.userId = userId;
		Spool.initSpool(CLIENTSSPOOLDIR + userId);
	}

	/**
	 * Envoi d'un message protocole PUT, et envoi du fichier désigné.
	 * 
	 * @param to       userId du destinataire du fichier/
	 * @param fileName nom du fichier à envoyer.
	 * @throws Exception Toutes les exceptions qui peuvent être levées pour la
	 *                   gestion de la connexion avec le serveur, et la lecture du
	 *                   fichier à envoyer.
	 */
	public void put(final String to, final String fileName) throws Exception {
		// on commence par ouvrir le fichier pour vérifier qu'il existe.
		FileInputStream fin = new FileInputStream(fileName);
		/*******************************************************************************/
		// Partie Protocole Interpreter: envoi PUT + réception RECEIVE
		// envoi du message de protcole PUT
		sendPutMessage(to, fileName);
		// réception du message de protcole SEND envoyé en réponse par le serveur.
		ProtocolMessage sendMessage = receiveProtocolMessage();
		if (sendMessage == null || sendMessage.getMessageType() != ProtocolMessageType.SEND) {
			throw new Exception(BADMESSAGE + " " + sendMessage + " " + sendMessage.getMessageType());
		}
		/*******************************************************************************/
		// Partie Data Transfert Process: connexion au DTP serveur + envoi du fichier

		// établissement avec la socket serveur préparé par le serveur pour l'échange du
		// contenu
		// du fichier.
		SocketChannel rwChan = SocketChannel.open(new InetSocketAddress(serverHost, sendMessage.getPort()));
		FileChannel fc = fin.getChannel();
		sendFile(fc, rwChan);
		fin.close();
		fc.close();
		// fermeture obligatoire de la connexion pour que le serveur sache que le
		// fichier est totalement envoyé.
		rwChan.close();
	}

	/**
	 * Envoi d'un message protocole GET et réception de tous les fichiers concernés
	 * en parallèle dans une boucle gérant les réceptions en mode asynchrone.
	 * 
	 * @throws Exception Toutes les exceptions qui peuvent être levées pour la
	 *                   gestion des connexions avec le serveur, et l'écriture des
	 *                   fichier à recevoir.
	 */
	public void get() throws Exception {
		// TODO Question 3
	}

	private void sendMessage(final ProtocolMessage message) throws IOException {
		worker.sendMsg(MessType.PROTOCOLMESSAGETYPE, message);
	}

	private void sendPutMessage(final String to, final String fileName) throws IOException {
		Information information = new Information(userId, to, fileName);
		ProtocolMessage putMessage = new ProtocolMessage(ProtocolMessageType.PUT, information);
		sendMessage(putMessage);
	}

	private void sendGetMessage() throws IOException {
		COMM.info("Client::sendGetMessage()");
		Information information = new Information(userId, null, null);
		ProtocolMessage putMessage = new ProtocolMessage(ProtocolMessageType.GET, information);
		sendMessage(putMessage);
	}

	private ProtocolMessage receiveProtocolMessage() throws IOException {
		ReadMessageStatus status = ReadMessageStatus.ReadUnstarted;

		while ((status != ReadMessageStatus.ChannelClosed) && (status != ReadMessageStatus.ReadDataCompleted)) {
			status = worker.readMessage();
		}

		if (status == ReadMessageStatus.ReadDataCompleted) {
			Serializable obj = worker.getData();
			if (obj instanceof ProtocolMessage) {
				return (ProtocolMessage) obj;
			}
		}
		return null;

	}
}
